UNIT matmath;

INTERFACE
 USES stdhdr;

PROCEDURE MatPrint( VAR m1: recmat; m,n: integer; matname: TitleStr);

PROCEDURE MatProd(VAR m1,m2: recmat; l,m,n: integer; VAR m3: recmat);

PROCEDURE MatScalarProd(VAR m1:recmat; sval:double; numrow,numcol:
                              integer; VAR m2:recmat);

PROCEDURE MatAdd(VAR m1,m2:recmat; numrow,numcol:integer; VAR m3:recmat);

PROCEDURE MatTranspose (VAR m1:recmat; numrow,numcol:integer; VAR m2:recmat);

FUNCTION MatDeter( m1:recmat; numrow:integer): double;

PROCEDURE MatInvert(VAR matdata:    recmat;
                         numcol:     integer;
                     VAR det:        double;
                     VAR invary:     recmat);

PROCEDURE NormalizeEigenVectors(VAR a: recmat; n: integer);

PROCEDURE CyclicJacobi(VAR a: recmat;
                       n: integer;
                  VAR eigenvalues: shortVector;
                  VAR v: recmat;
                  VAR  count: integer;
                  VAR  success: boolean);

procedure amal21(var phi  :recmat;
                 var y    :longvector;
                 var l,m  :integer;
                 var coef : longvector);

IMPLEMENTATION

PROCEDURE MatPrint( VAR m1: recmat; m,n: integer; matname: TitleStr);

VAR i,j: integer;

begin
  writeln(matname);
  writeln;
  for i := 0 to m-1  do
  begin
    for j:= 0 to n-1  do
      write(m1[i,j]:10:3,'  ');
    writeln;
  end;
  writeln;
end;

PROCEDURE MatProd(VAR m1,m2: recmat; l,m,n: integer; VAR m3: recmat);
VAR
    i,j,k: integer;

begin   {PROCEDURE MatProd}
  for i:= 0 to l-1 do
    for j:= 0 to n-1 do
    begin
      m3[i,j] := 0;
      for k:= 0 to m-1 do
        m3[i,j] := m3[i,j] + m1[i,k] * m2[k,j];
    end;
end;

PROCEDURE MatScalarProd(VAR m1:recmat; sval:double; numrow,numcol:
                              integer; VAR m2:recmat);

VAR i,j: integer;

begin  {PROCEDURE MatScalarProd}
  for i:= 0 to numrow-1  do
    for j:= 0 to numcol-1 do
      m2[i,j] := sval * m1[i,j];
end;

PROCEDURE MatAdd(VAR m1,m2:recmat; numrow,numcol:integer; VAR m3:recmat);

VAR i,j: integer;

begin {PROCEDURE MatAdd}
  for i:= 0 to numrow-1  do
    for j:= 0 to numcol-1 do
      m3[i,j] := m1[i,j] + m2[i,j];
end;

PROCEDURE MatTranspose (VAR m1:recmat; numrow,numcol:integer; VAR m2:recmat);

VAR i,j: integer;

begin {PROCEDURE MatTranspose}
  for i:= 0 to numrow-1  do
    for j:= 0 to numcol-1 do
      m2[j,i] := m1[i,j];
end;



function MatDeter( m1:recmat; numrow:integer): double;

VAR i,j,k,lx,f,nxt:                      integer;
    pivot,rac,bignum,temp1,temp2,eval:  double;

begin  {PROCEDURE MatDeter}
  f := 1;
  for i:= 0 to numrow-1 do
  begin
    bignum := 0;
    for k := i to numrow-1 do
    begin
      eval := abs(m1[k,i]);
      if eval - bignum > 0 then
      begin
        bignum := eval;
        lx := k;
      end;
    end;
    if i - lx <> 0 then
      f := -f;
    for j := 0 to numrow-1  do
    begin
      temp1 := m1[i,j];
      m1[i,j] := m1[lx,j];
      m1[lx,j] := temp1;
    end;
    pivot := m1[i,i];
    nxt := i + 1;
    for j := nxt to numrow-1 do
    begin
      rac := m1[j,i]/pivot;
      for k := i to numrow-1  do
        m1[j,k] := m1[j,k] - rac * m1[i,k];
    end;
  end;
  temp2 := 1;
  for i := 0 to numrow-1 do
     temp2 := temp2 * m1[i,i];
  MatDeter := temp2 * f;
end;

PROCEDURE MatInvert(VAR matdata:    recmat;
                         numcol:     integer;
                     VAR det:        double;
                     VAR invary:     recmat);

VAR
  pivlst:                 ARRAY[0..maxc,0..1] OF integer;
  pivchk:                 ARRAY[0..maxc] OF boolean;
  i,j,k,l,lerow,lecol,l1: integer;
  piv,t,leval:            double;


PROCEDURE swap(VAR s1,
                   s2: double);
VAR
  temp: double;

BEGIN         {PROCEDURE swap}
  temp := s1;
  s1 := s2;
  s2 := temp;
END;



BEGIN {PROCEDURE MatInvert}
det := 1;
FOR i := 0 TO numcol-1 DO
  BEGIN
    pivchk[i] := false;
    FOR j := 0 TO numcol-1 DO
      invary[i,j] := matdata[i,j];
  END;

  FOR i:= 0 TO numcol-1 DO
  BEGIN
    leval := 0.0;
    FOR j := 0 TO numcol-1 DO
       IF NOT(pivchk[j]) THEN
         FOR k := 0 TO numcol-1 DO
           IF NOT(pivchk[k])  THEN
             IF abs(invary[j,k]) > leval THEN
             BEGIN
               lerow := j;
               lecol := k;
               leval := abs(invary[j,k])
             END;

     pivchk[lecol] := true;
     pivlst[i,0] := lerow;
     pivlst[i,1] := lecol;

     IF lerow <> lecol THEN
     BEGIN
        det := -det;
        FOR l := 0 TO numcol-1 DO
          swap(invary[lerow,l],invary[lecol,l]);
     END;

     piv := invary[lecol,lecol];
     det := det * piv;
     IF det > 1.0e+30 THEN det := 1;      {prevents numeric overflow}
     invary[lecol,lecol] := 1.0;
     FOR l := 0 TO numcol-1 DO
       invary[lecol,l] := invary[lecol,l]/piv;

     FOR l1 := 0 TO numcol-1 DO
     BEGIN
       IF l1 <> lecol THEN
       BEGIN
         t := invary[l1,lecol];
         invary[l1,lecol] := 0;
         FOR l := 0 TO numcol-1 DO
           invary[l1,l] := invary[l1,l] - invary[lecol,l] * t;
       END
     END
   END;

 FOR i := 0 TO numcol-1 DO
 BEGIN
   l := numcol - i - 1;
   IF pivlst[l,0] <> pivlst[l,1] THEN
   BEGIN
     lerow := pivlst[l,0];
     lecol := pivlst[l,1];
     FOR k := 0 TO numcol-1 DO
       swap(invary[k,lerow],invary[k,lecol])
   END
 END;
END;


PROCEDURE NormalizeEigenVectors(VAR a: recmat; n: integer);
VAR i,j: integer;
    max: double;
begin
  for j := 0 to n-1 do
  begin
    max := a[0,j];
    for i := 1 to n-1 do
      if abs(a[i,j]) > abs(max) then max := a[i,j];
    for i := 0 to n-1 do
      a[i,j] := a[i,j]/max;
  end;
end;


PROCEDURE CyclicJacobi(VAR a: recmat;
                       n: integer;
                  VAR eigenvalues: shortVector;
                  VAR v: recmat;
                  VAR  count: integer;
                  VAR  success: boolean);


const nearzero = 1.0e-8;  {change to 1.0e-15 for 8087 system}
      tolerance = 1.0e-8;

VAR maxCount, i,p,q : integer;
    PiOver4, PiOver2, sumsq: double;
    avp, avq: shortvector;

PROCEDURE JacobiRotation(VAR m: recmat;
                 p,q: integer;
            VAR mp, mq: shortvector);

VAR  g,h,mpp, mqq, mpq, mpj, mqj,
     tau,tempr, angle, c,s,cs,cSquared,sSquared: double;
     j: integer;

begin
  mpp := mp[p]; mpq := mp[q]; mqq := mq[q];
  if abs(mpp-mqq) > nearzero then
  begin
    angle := arctan(2.0 * mpq/(mpp-mqq))/2.0;
    if angle > PiOver4 then
      angle := angle - PiOver2
  end else
    if mpq >0 then angle := PiOver4 else angle := -PiOver4;

  c := cos(angle) ; s := sin(angle);
  cSquared := c*c; sSquared := s*s ; cs := c * s;
  tempr := 2.0 * mpq * cs;
  mp[p] := mpp * cSquared + tempr + mqq * sSquared;
  mq[q] := mpp * sSquared - tempr + mqq * cSquared;
  mp[q] := 0.0 ; mq[p] := 0.0;
  tau := s / (1.0 + c);
  for j := 0 to n-1 do
    if not(j in [p,q] ) then
    begin
      mpj := mp[j] * c + mq[j] * s;
      mqj := mq[j] * c - mp[j] * s;
      mp[j] := mpj; m[j,p] := mpj;
      mq[j] := mqj; m[j,q] := mqj;
    end;
    for j := 0 to n-1 do
    begin
      g := v[j,p]; h := v[j,q];
      v[j,p] := c * g  + s * h;
      v[j,q] := -s * g +  c * h;
    end;

end;

begin
  maxCount :=  count;
  count := 0;
  for p := 0 to n-1 do
  begin
    for q := 0 to n-1 do
      v[p,q] := 0.0;
    v[p,p] := 1.0;
  end;

  PiOver4 := pi/4.0; PiOver2 := pi/2.0;
  repeat
    count := count + 1; 
    sumsq := 0.0;
    for p := 0 to n-1 do
      sumsq := sumsq + a[p,p] * a[p,p];
    success := true;
    for p := 0 to n -2  do
      for q := p + 1 to n-1 do
        if abs(a[p,q]/sumsq) > tolerance then
        begin
          success := false;
          JacobiRotation(a,p,q,a[p],a[q]);
        end;
  until success or (count > maxCount);
 if success then
 begin
    for p := 0 to n-1 do
      eigenvalues[p] := a[p,p];

  end;
end;

{moindres carrs linaires}
//l:=nombre de lignes = nbre d'quations;
//m:= nombre de colonnes = nbre de parametres;
procedure amal21(var phi  : recmat;
                 var y    : longvector;
                 var l,m  : integer;
                 var coef : longvector);{
-----------------------------------------}
var
  phit,aux,phiphitm1,prod                   : recmat;
  i,j                                       : integer;
  determinant,s                             : double;
begin
  MatTranspose(phi,l,m,phit);
  MatProd(phi,phit,l,m,l,aux);
  MatInvert(aux,l,determinant,phiphitm1);
  MatProd(phiphitm1,phi,l,l,m,prod);
  for i:=0 to l-1 do begin
    s:=0;
    for j:=0 to m-1 do s:=s+y[j]*prod[i,j];
    coef[i]:=s;
  end;
end;

END.
